use crate::{AccountId, RevId, Tag, TxId, Type};
use chrono::{DateTime, Utc};
use serde::{Deserialize, Serialize};

/// Filter transactions
///
/// All this filters options are AND, meaning that an object must match all the
/// requirements in order to be included in the resultset.
#[derive(Clone, Debug, Serialize, Deserialize, Default)]
pub struct Filter {
    /// List of transaction IDs to query
    #[serde(skip_serializing_if = "Vec::is_empty")]
    pub ids: Vec<TxId>,
    /// List of revisions to query
    #[serde(skip_serializing_if = "Vec::is_empty")]
    pub revisions: Vec<RevId>,
    /// List of accounts to query their transactions
    #[serde(skip_serializing_if = "Vec::is_empty")]
    pub accounts: Vec<AccountId>,
    /// List of transaction types-kind
    #[serde(rename = "type", skip_serializing_if = "Vec::is_empty")]
    pub typ: Vec<Type>,
    /// List of transactions by tags
    #[serde(skip_serializing_if = "Vec::is_empty")]
    pub tags: Vec<Tag>,
    /// List transactions newer than this timestamp
    #[serde(
        default,
        with = "option_ts_seconds",
        skip_serializing_if = "Option::is_none"
    )]
    pub since: Option<DateTime<Utc>>,
    /// List transactions upto this timestamp
    #[serde(
        default,
        with = "option_ts_seconds",
        skip_serializing_if = "Option::is_none"
    )]
    pub until: Option<DateTime<Utc>>,
    /// Limit for transactions
    #[serde(default, skip_serializing_if = "is_zero")]
    pub limit: usize,
    /// Skip the first `skip` transactions
    #[serde(default, skip_serializing_if = "is_zero")]
    pub skip: usize,
}

impl Filter {
    /// Adds a given kind to the filter
    pub fn kind(mut self, kind: Type) -> Self {
        self.typ.push(kind);
        self
    }

    /// Adds a given tag to the filter
    pub fn account(mut self, account: AccountId) -> Self {
        self.accounts.push(account);
        self
    }

    /// Adds until timestamp to the filter
    pub fn until(mut self, time: DateTime<Utc>) -> Self {
        self.until = Some(time);
        self
    }

    /// Adds since timestamp to the filter
    pub fn since(mut self, time: DateTime<Utc>) -> Self {
        self.since = Some(time);
        self
    }
}

impl From<Vec<Tag>> for Filter {
    fn from(value: Vec<Tag>) -> Self {
        Filter {
            tags: value,
            ..Default::default()
        }
    }
}

impl From<TxId> for Filter {
    fn from(value: TxId) -> Self {
        Filter {
            ids: vec![value],
            ..Default::default()
        }
    }
}

impl From<RevId> for Filter {
    fn from(value: RevId) -> Self {
        Filter {
            revisions: vec![value],
            ..Default::default()
        }
    }
}

impl From<AccountId> for Filter {
    fn from(value: AccountId) -> Self {
        Filter {
            accounts: vec![value],
            ..Default::default()
        }
    }
}

#[inline]
fn is_zero(number: &usize) -> bool {
    *number == 0
}

pub(crate) mod option_ts_seconds {
    use chrono::{DateTime, LocalResult, TimeZone, Utc};
    use serde::{de, Deserialize, Deserializer, Serializer};

    pub fn serialize<S>(date: &Option<DateTime<Utc>>, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: Serializer,
    {
        match date {
            Some(date) => serializer.serialize_i64(date.timestamp()),
            None => serializer.serialize_none(),
        }
    }

    pub fn deserialize<'de, D>(deserializer: D) -> Result<Option<DateTime<Utc>>, D::Error>
    where
        D: Deserializer<'de>,
    {
        let timestamp: Option<i64> = Deserialize::deserialize(deserializer)?;
        let timestamp = if let Some(timestamp) = timestamp {
            timestamp
        } else {
            return Ok(None);
        };

        match Utc.timestamp_opt(timestamp, 0) {
            LocalResult::None => Err(de::Error::custom("Invalid timestamp")),
            LocalResult::Ambiguous(min, max) => Err(de::Error::custom(format!(
                "Timestamp {0} is ambiguous (min={1}, max={2})",
                timestamp, min, max
            ))),
            LocalResult::Single(ts) => Ok(Some(ts)),
        }
    }
}
