//! Cursor implementation
use crate::{AccountId, BaseTx, Filter, RevId, Revision, Tag, TxId, Type};

#[derive(Debug, Clone)]
/// The primary filter is used to filter the transactions before applying the other filters. Think
/// of it like which key is used to filter transactions quickly before applying the other filters.
pub enum PrimaryFilter {
    /// By transaction ID
    Id(Vec<TxId>),
    /// By revision ID
    Revision(Vec<RevId>),
    /// By accounts
    Account(Vec<AccountId>),
    /// By transaction type
    Type(Vec<Type>),
    /// By tags
    Tags(Vec<Tag>),
    /// By transaction status
    Stream,
}

#[derive(Debug, Clone)]
/// The cursor
pub struct Cursor {
    /// The primary filter
    pub primary_filter: PrimaryFilter,
    /// The secondary filter
    pub filter: Filter,
}

impl From<Filter> for Cursor {
    fn from(mut filter: Filter) -> Self {
        let primary_filter = if !filter.revisions.is_empty() {
            PrimaryFilter::Revision(std::mem::take(&mut filter.revisions))
        } else if !filter.ids.is_empty() {
            PrimaryFilter::Id(std::mem::take(&mut filter.ids))
        } else if !filter.accounts.is_empty() {
            PrimaryFilter::Account(std::mem::take(&mut filter.accounts))
        } else if !filter.typ.is_empty() {
            PrimaryFilter::Type(std::mem::take(&mut filter.typ))
        } else if !filter.tags.is_empty() {
            PrimaryFilter::Tags(std::mem::take(&mut filter.tags))
        } else {
            PrimaryFilter::Stream
        };

        filter.ids.sort();
        filter.revisions.sort();
        filter.accounts.sort();
        filter.typ.sort();
        filter.tags.sort();

        Self {
            primary_filter,
            filter,
        }
    }
}

impl Cursor {
    /// Check if the base transaction and the revision matches the current cursor filter
    pub fn matches(&self, base: &BaseTx, revision: &Revision) -> bool {
        if !self.filter.ids.is_empty()
            && self
                .filter
                .ids
                .binary_search(&revision.transaction_id)
                .is_err()
        {
            return false;
        }

        if !self.filter.revisions.is_empty()
            && self
                .filter
                .revisions
                .binary_search(&revision.rev_id().expect("vv"))
                .is_err()
        {
            return false;
        }

        if !self.filter.typ.is_empty() && self.filter.typ.binary_search(&base.typ).is_err() {
            return false;
        }

        if !self.filter.tags.is_empty() {
            let mut found = false;
            for tag in revision.tags.iter() {
                if self.filter.tags.binary_search(&tag).is_ok() {
                    found = true;
                    break;
                }
            }
            if !found {
                return false;
            }
        }

        true
    }
}
